/*-----------------------------------------------------------------
  Thermometer1 Copyright (c) 2007 by Jack Botner

  main.c

  This program implements a thermometer using a 3 digit LED display.
  Either a LM34 or a TC1047 temperature sensor may be used. Currently
  negative temperatures are not implemented.

  Processor: ATMega8
      Clock: External, 3.579545 MHz

-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "led.h"

#define INIT_CTR0   144				// 256 - 112
#define INIT_CTR1   20797			// 65536 - 44739

#define EVENT_ADC_COMPLETE	0		// adc complete event flags bit position
#define DIGITAL_FILTER_SAMPLES 25	// to reduce jitter in the display

// Global variables

static char cDisplay[3] = { 0, 0, 0 };	// digits to display

static char cEvents=0;				// event flags

static uint16_t	uiAdc=0;			// last adc sample
static uint16_t	uiSmoothedAdc=0;	// smoothed adc reading
static uint16_t	uiTempTenths=0;		// temperature in 10ths of degrees

// Internal Functions

//static void formatTime( void );
static void calculate_temperature(void);
static unsigned int calculate_temperature_lm34( void );
static unsigned int calculate_temperature_tc1047( void );
static void digital_filter(void);
static void display_temperature(void);
static void system_init(void);

/*-----------------------------------------------------------------
  main()

  When an adc reading is available, processes the reading into
  temperature and display on the LEDs. All other work done in
  the interrupt handling routines.
-----------------------------------------------------------------*/
int main(void)
{
  system_init();			// initialise system

  while( 1 )
  {
    if ( test_bit( cEvents, EVENT_ADC_COMPLETE ) )
	{
	  digital_filter();				// filter the adc output
	  calculate_temperature();		// convert adc to temperature
	  display_temperature();		// display temperature on the LEDs
	  clear_bit( cEvents, EVENT_ADC_COMPLETE );
    }
	
	_delay_loop_2( 8948 );		// 10 ms
  }
}

/*-----------------------------------------------------------------
  Timer 0 overflow interrupt handler - updates LED display

  To achieve an interval of approximately 2 ms for selecting
  the next LED digit. By sequencing each digit fast enough, the LED
  digits appear to be all on.

  Since the clock is running at 3.579545 mHz, the counter increments
  every 279.4 ns. Using a count of 112 and prescaler of 64 we get
  an interrupt interval of 2.003 ms.
-----------------------------------------------------------------*/
ISR(TIMER0_OVF_vect)
{
  static unsigned char ucIndex=0;

  ledSelectDigit( NO_DIGIT );               // turn all digits off

  ledOutputNumber( cDisplay[ucIndex] );		// output the number

  ledSelectDigit( ucIndex );                // select the digit

  ++ucIndex;
  if ( ucIndex == 3 ) ucIndex = 0;

  TCNT0 = INIT_CTR0;						// reinitialize counter
}

/*-----------------------------------------------------------------
  Timer 1 overflow interrupt handler - initiates adc conversion

  To achieve an interval of approximately 300 ms for initiating
  the next adc conversion. This pretty much dictates how often/fast
  the display is updated.

  Since the clock is running at 3.579545 mHz, the counter increments
  every 279.4 ns. Using a count of 44739 and prescaler of 8 we get
  an interrupt interval of 100 ms.
-----------------------------------------------------------------*/
ISR(TIMER1_OVF_vect)
{
  static uint8_t ctr=0;

  if ( ctr < 3 )
    ++ctr;
  else
  {
    set_bit( ADCSRA, ADSC );		// initiate a conversion
	ctr = 0;
  }

  TCNT1 = INIT_CTR1;				// reinitialize counter
}

/*-----------------------------------------------------------------
  ISR( ADC_vect ) - ADC conversion complete interrupt handler

  ADC0 is monitoring the temperature sensor.
-----------------------------------------------------------------*/
ISR( ADC_vect )
{
  // ADCL must be read before ADCH.
  uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );

  set_bit( cEvents, EVENT_ADC_COMPLETE );
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (output) LED digit 1 select
         PB1 (output) LED digit 2 select
         PB2 (output) LED digit 3 select
         PB3 (n/a) MOSI (ISP)
         PB4 (n/a) MISO (ISP)
         PB5 (n/a) SCK (ISP)
         PB6 (n/a) XTAL1
         PB7 (n/a) XTAL2

  PORTC: PC0 (input) ADC0 Temperature sensor 
         PC1 (input) ADC1 (reserved) 
         PC2 (input) Sensor type signal
         PC3 (input) (reserved)
         PC4 (input) J3 jumper (C or F)
         PC5 (output) not used
         PC6 (n/a) !Reset (ISP)

  PORTD: PD0 (output) LED decoder B
         PD1 (output) LED decoder C
         PD2 (output) LED decoder D
         PD3 (output) LED decoder A
         PD4 (output) LED DP
         PD5 (output) not used
         PD6 (output) not used
         PD7 (output) not used

-----------------------------------------------------------------*/
static void system_init()
{
  DDRB  = 0b00000111;       // PortB0..2 = output, PortB3..7 = input
  PORTB = 0b00000000;       // initialise PortB7

  DDRC  = 0b00100000;       // PortC0..4 = input, PortC5 = output
  PORTC = 0b00010000;       // initialise PortC7

  DDRD  = 0b11111111;       // PortD0..7 = output
  PORTD = 0b00000000;       // initialise PortD7

  // Initialize Timer0. Timer0 is used to output digits on the LED
  // display every 2 milliseconds.
  TCCR0 = 0b00000011;       // prescaler = 64 (turns timer0 on)
  TCNT0 = INIT_CTR0;
  set_bit( TIMSK, TOIE0 );	// enable timer/counter0 overflow interrupt

  // Initialize Timer1. Timer1 is used to initiate adc conversions
  // every 300 milliseconds.

  TCCR1A = 0b00000000;		// normal operation
  TCCR1B = 0b00000010;		// prescaler = 8 (turns timer1 on)
  TCNT1 = INIT_CTR1;
  set_bit( TIMSK, TOIE1 );	// enable timer/counter1 overflow interrupt

  // Set up PC0/ADC0 as analog input, Vref=Aref
  ADMUX  = 0b11000000;		// Vref=2.56V; result right adjusted; select ADC0
  ADCSRA = 0b10001101;		// Enable ADC; Enable interrupt; prescaler = 32

  sei();                    // enable global interrupts
}

/*---------------------------------------------------------------------
  digital_filter()

  Why do we need this? Because the output of the adc jitters one or
  two counts each sample, which translates to one or two tenths of a
  degree C on the display. It goes up and down apparently randomly,
  which is very annoying. According to the Atmel spec, the adc has a
  number of error modes which result in the lsb being uncertain.
---------------------------------------------------------------------*/
static void digital_filter()
{
  static uint16_t uiTable[DIGITAL_FILTER_SAMPLES];
  static uint8_t  ucIndex=0, ucEntries=0;
  uint8_t		  ii;
  uint32_t		  ulSum;

  // Add the entry to the table, wrap-around
  uiTable[ucIndex] = uiAdc;
  ++ucIndex;
  if ( ucIndex >= DIGITAL_FILTER_SAMPLES )
    ucIndex = 0;
  if ( ucEntries < DIGITAL_FILTER_SAMPLES )
    ++ucEntries;

  // Average the table
  ulSum = 0;
  for ( ii=0; ii<ucEntries; ++ii )
    ulSum += uiTable[ii];

  uiSmoothedAdc = (unsigned int) ( ulSum / ucEntries );
}

/*---------------------------------------------------------------------
  calculate_temperature()

  Calculate the temperature from the adc output.
  Accept either LM34 or TC1047 data (J1).
  Output in degrees C or F as required by J3.
---------------------------------------------------------------------*/
static void calculate_temperature()
{
  if ( test_bit( PINC, PC2 ) )			// test J1 
    uiTempTenths = calculate_temperature_lm34();
  else
    uiTempTenths = calculate_temperature_tc1047();

  //uiTempTenths = uiSmoothedAdc;		// testing!
}

/*---------------------------------------------------------------------
  calculate_temperature_lm34()

  Convert the A/D input value (0..1023 representing a voltage in the
  range 0..2.56 volts) to tenths of degrees C or F.
---------------------------------------------------------------------*/
static unsigned int calculate_temperature_lm34()
{
  unsigned long ulADCmv, ulTempC;

  // uiAdc0 (adresh:adresl) is in the range 0..1023 which represents a 
  // voltage of 0..2.56 volts (Vref). Scale the adc value to a voltage value.

  ulADCmv = ( 2560L * (unsigned long) uiSmoothedAdc ) / 1024L;
  
  // The A/D voltage (eg. 750 mv) represents 75.0 degrees F
  // Keep in mind this is 10 times the actual temperature
  
  if ( !test_bit( PINC, PC4 )	)		// J3 want degrees F
    return (unsigned int) ulADCmv;

  // Convert Farenheit to Celsius
  
  ulTempC = ( ( ulADCmv - 320L ) * 5L ) / 9L;

  return (unsigned int) ulTempC;
}

/*---------------------------------------------------------------------
  calculate_temperature_tc1047()

  Convert the A/D input value (0..1023 representing a voltage in the
  range 0..2.56 volts) to tenths of degrees C. Vref = 2.56 V.
  
  First we need to convert the A/D voltage to millivolts, then
  to tenths of degrees C.
  
  The formula for conversion is 
  
				  V - 500
			T = ------------
				   10.0
				   
  where T is temperature in degrees C, and V is Vout in millivolts.
  
  Note: we're not going to handle negatives for now!
  
    Input: A/D reading
  Returns: tenths of degrees C or F
---------------------------------------------------------------------*/
static unsigned int calculate_temperature_tc1047()
{
  unsigned long ulVoutMv, ulTempF;
  unsigned int uiTempTensC;

  // Input is in the range 0..1023 which represents a voltage of 0..2.56
  // volts (Vref). Scale the adc value to a voltage value (millivolts).

  ulVoutMv = ( (unsigned long) uiSmoothedAdc * 2560L ) / 1024L;

  // Derive temperature (tens of degrees C) from the formula above.
  // Multiply numerator and denominator by 100 to handle decimals.
  // Multiply numerator by 10 to get tens of degrees.
  
  //uiTempTensC = (unsigned int) ( ( ( ulVoutMv - 500L ) * 1000L ) / 1000L );
  uiTempTensC = (unsigned int) ( ulVoutMv - 500L );

  if ( test_bit( PINC, PC4 )	)		// J3 want degrees C
    return uiTempTensC;

  // Convert Celsius to Fahrenheit

  ulTempF = ( (unsigned long) uiTempTensC * 9L / 5L ) + 320L;

  return (unsigned int) ulTempF;
}

/*---------------------------------------------------------------------
  display_temperature()

  Convert the temperature in 10ths degrees to displayable format,
  and load resuly in cDisplay array.
---------------------------------------------------------------------*/
static void display_temperature()
{
  char  buffer[6];
  char  nn;

  utoa( uiTempTenths, buffer, 10 );
  nn = (char) strlen( buffer );
  if ( !nn )
  {
    cDisplay[0] = 0x0f;
    cDisplay[1] = 0x0f;
    cDisplay[2] = '0';
  }
  else
    if ( nn == 1 )
    {
      cDisplay[0] = 0x0f;
      cDisplay[1] = 0x0f;
      cDisplay[2] = buffer[0];
    }
    else
      if ( nn == 2 )
      {
        cDisplay[0] = 0x0f;
        cDisplay[1] = buffer[0];
        cDisplay[2] = buffer[1];
      }
	  else
        if ( nn == 3 )
        {
          cDisplay[0] = buffer[0];
          cDisplay[1] = buffer[1];
          cDisplay[2] = buffer[2];
        }
	    else
        {
          cDisplay[0] = '9';
          cDisplay[1] = '9';
          cDisplay[2] = '9';
        }
}
